@@ -101,6 +101,7 @@ def exec_ui(
         use_script_app: bool = False,
         icons: Tuple[str | Path, ...] = None,
         immediate_run: bool = False,
+        allow_kernel_interrupt: bool = False,
 ):
     """
     Decorates the function as an Exec UI function. We have different kinds of UI functions. By default,
@@ -136,6 +137,7 @@ def exec_ui(
         wrapper.__ui_input_request__ = input_request
         wrapper.__ui_immediate_run__ = immediate_run
         wrapper.__ui_icons__ = icons
+        wrapper.__ui_allow_kernel_interrupt__ = allow_kernel_interrupt
         if use_script_app:
             wrapper.__ui_runnable__ = RUNNABLE_SCRIPT
         elif use_kernel:
