@@ -6,10 +6,10 @@ import jupyter_client.kernelspec
 import rich
 from executor import ExternalCommand
 from executor import ExternalCommandFailed
+from jupyter_client import KernelClient
 from jupyter_client.manager import KernelManager
 from rich import print
-
-from gui_executor.utils import remove_ansi_escape
+from rich.console import Console
 
 LOGGER = logging.getLogger("gui-executor.kernel")
 
@@ -18,46 +18,51 @@ class KernelError(Exception):
     pass
 
 
-def decode_traceback(traceback: str) -> str:
-    return remove_ansi_escape('\n'.join(traceback))
+def find_running_kernels():
+    import psutil, datetime
+    all_python = [proc for proc in psutil.process_iter() if "python" in proc.name().lower()]
+    kernel_processes = [y for y in all_python for z in y.cmdline() if 'ipykernel' in z]
+    datetime.datetime.fromtimestamp(kernel_processes[0].create_time()).strftime("%Y-%m-%d %H:%M:%S")
 
+# TODO:
+#    * Create a separate kernel and client class
+#      For the kernel class, check if there is a need to re-connect to an existing kernel first.
+#      For the client class, connect to the kernel with:
+#
+#        from jupyter_client import BlockingKernelClient
+#
+#        client = BlockingKernelClient()
+#        client.load_connection_file('/Users/ebanner/Library/Jupyter/runtime/kernel-10962.json')
+#        client.start_channels()
+#
+#    * see also
+#      https://ipython.readthedocs.io/en/stable/development/wrapperkernels.html
+#      https://github.com/ipython/ipykernel
 
 class MyKernel:
 
     def __init__(self, name: str = "python3", startup_timeout: int = 60):
         self._kernel = KernelManager(kernel_name=name)
         self._kernel.start_kernel()
-        self._client = self._kernel.client()
-        self._client.start_channels()
-        try:
-            self._client.wait_for_ready(timeout=startup_timeout)
-        except RuntimeError:
-            self._client.stop_channels()
-            self._kernel.shutdown_kernel()
-            raise
+        self._startup_timeout = startup_timeout
         self._error = None
 
     def is_alive(self) -> bool:
         return self._kernel.is_alive()
 
     def shutdown(self):
-        self._client.stop_channels()
         self._kernel.shutdown_kernel(now=True)
 
     def interrupt_kernel(self):
         self._kernel.interrupt_kernel()
 
+    def get_client(self) -> KernelClient:
+        return self._kernel.client()
+
     @staticmethod
     def get_kernel_specs():
         return jupyter_client.kernelspec.find_kernel_specs()
 
-    def get_kernel_info(self):
-        msg_id = self._client.kernel_info()
-        # rich.print(f"get_info() -> {msg_id = }")
-        reply = self._client.get_shell_msg()
-        # rich.print("get_info() -> reply = ", reply)
-        return reply
-
     def get_connection_file(self):
         return self._kernel.connection_file
 
@@ -71,73 +76,6 @@ class MyKernel:
         )
         return info
 
-    @property
-    def client(self):
-        return self._client
-
-    def run_snippet(self, snippet: str) -> str:
-
-        # Flush the IOPub channel before executing the command. This is needed because another
-        # client might be connected that has sent out messages on the pub channel. We want to
-        # catch the output of the given command of course.
-
-        self.flush()
-
-        msg_id = self._client.execute(snippet)
-        LOGGER.debug(f"{msg_id = }")
-
-        reply = self._get_response(msg_id)
-        LOGGER.debug(f"{type(reply) = }")
-        LOGGER.debug(f"{reply = }")
-        LOGGER.debug(f"{reply['content'] = }")
-
-        if reply["content"]["status"] == "error":
-            try:
-                self._error = decode_traceback(reply["content"]["traceback"])
-            except KeyError:
-                self._error = "An error occurred, no traceback was provided."
-        else:
-            self._error = None
-
-        output = self._get_output()
-        # LOGGER.debug(f"{output = }")
-
-        return output.strip()
-
-    def flush(self):
-        while True:
-            try:
-                _ = self._client.get_iopub_msg(timeout=0.2)
-            except queue.Empty:
-                break
-
-    def get_error(self) -> str:
-        return self._error
-
-    def _get_response(self, msg_id: str = None) -> dict:
-        return self._client.get_shell_msg(msg_id)
-
-    def _get_output(self, timeout: float = 1.0) -> str:
-
-        # When the execution state is "idle" it is complete
-
-        io_msg_content = self._client.get_iopub_msg(timeout=timeout)['content']
-
-        if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
-            return ""
-
-        while True:
-            last_io_msg_content = io_msg_content
-
-            try:
-                io_msg_content = self._client.get_iopub_msg(timeout=timeout)['content']
-                if 'execution_state' in io_msg_content and io_msg_content['execution_state'] == 'idle':
-                    break
-            except queue.Empty:
-                break
-
-        return self._decode_io_msg_content(last_io_msg_content)
-
     @staticmethod
     def _decode_io_msg_content(content: dict) -> str:
 
@@ -152,9 +90,18 @@ class MyKernel:
 
 
 def do_test_my_kernel(name: str = "python3"):
+    from gui_executor.client import MyClient
+    from gui_executor.utils import Timer
 
     kernel = MyKernel(name=name)
 
+    rich.print(kernel.get_kernel_specs())
+
+    client = MyClient(kernel)
+    client.connect()
+
+    rich.print(client.get_kernel_info())
+
     snippets = [
         "a=2",
         textwrap.dedent("""\
@@ -173,9 +120,16 @@ def do_test_my_kernel(name: str = "python3"):
         '!pip list -v'
     ]
 
+    console = Console(width=240)
     for snippet in snippets:
-        if out := kernel.run_snippet(snippet):
-            print(out)
+        if out := client.run_snippet(snippet):
+            console.print(out)
+
+    client.disconnect()
+
+    with Timer("MyClient as context manager"):
+        with MyClient(kernel) as client:
+            info = client.get_kernel_info()
 
 
 def start_qtconsole(kernel: MyKernel,
